package de.unima.ki.anyburl;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Properties;

import de.unima.ki.anyburl.algorithm.PathSampler;
import de.unima.ki.anyburl.data.Triple;
import de.unima.ki.anyburl.data.TripleSet;
import de.unima.ki.anyburl.io.IOHelper;
import de.unima.ki.anyburl.structure.Path;
import de.unima.ki.anyburl.structure.Rule;
import de.unima.ki.anyburl.threads.Predictor;
import de.unima.ki.anyburl.threads.Scorer;

public class Learn {
	
	private static long timeStamp = 0;
	
	private static String CONFIG_FILE = "config-learn.properties";

	
	
	/**
	 * Path to the file that contains the triple set used for learning the rules.
	 */
	public static String[] SINGLE_RELATIONS = null;
	
	private static LinkedList<Rule> rulesToBeScored = new LinkedList<Rule>();
	private static LinkedList<Rule> rulesScored = new LinkedList<Rule>();
	private static boolean finished = false;


	public static void main(String[] args) throws FileNotFoundException {
		
		if (args.length == 1) {
			CONFIG_FILE = args[0];
			System.out.println("reading params from file " + CONFIG_FILE);
		}
		
		Properties prop = new Properties();
		InputStream input = null;
		
		try {
			input = new FileInputStream(CONFIG_FILE);
			prop.load(input);
			SINGLE_RELATIONS = IOHelper.getProperty(prop, "SINGLE_RELATIONS", SINGLE_RELATIONS);
			Settings.PATH_TRAINING = IOHelper.getProperty(prop, "PATH_TRAINING",Settings.PATH_TRAINING);
			Settings.PATH_OUTPUT = IOHelper.getProperty(prop, "PATH_OUTPUT", Settings.PATH_OUTPUT);	
			Settings.SNAPSHOTS_AT = IOHelper.getProperty(prop, "SNAPSHOTS_AT", Settings.SNAPSHOTS_AT);
			Settings.SAMPLE_SIZE = IOHelper.getProperty(prop, "SAMPLE_SIZE", Settings.SAMPLE_SIZE);
			Settings.TRIAL_SIZE = IOHelper.getProperty(prop, "TRIAL_SIZE", Settings.TRIAL_SIZE);
			Settings.BATCH_TIME = IOHelper.getProperty(prop, "BATCH_TIME", Settings.BATCH_TIME);
			Settings.SATURATION = IOHelper.getProperty(prop, "SATURATION", Settings.SATURATION);
			Settings.WORKER_THREADS = IOHelper.getProperty(prop, "WORKER_THREADS", Settings.WORKER_THREADS);
			Settings.MAX_LENGTH_CYCLIC = IOHelper.getProperty(prop, "MAX_LENGTH_CYCLIC", Settings.MAX_LENGTH_CYCLIC);
			Settings.MAX_LENGTH_ACYCLIC = IOHelper.getProperty(prop, "MAX_LENGTH_ACYCLIC", Settings.MAX_LENGTH_ACYCLIC);
			Settings.THRESHOLD_CORRECT_PREDICTIONS = IOHelper.getProperty(prop, "THRESHOLD_CORRECT_PREDICTIONS", Settings.THRESHOLD_CORRECT_PREDICTIONS);
			Settings.THRESHOLD_CONFIDENCE = IOHelper.getProperty(prop, "THRESHOLD_CONFIDENCE", Settings.THRESHOLD_CONFIDENCE);
			
		}
		catch (IOException ex) {
			System.err.println("Could not read relevant parameters from the config file " + CONFIG_FILE);
			ex.printStackTrace();
			System.exit(1);
		}
		
		finally {
			if (input != null) {
				try {
					input.close();
				}
				catch (IOException e) {
					e.printStackTrace();
					System.exit(1);
				}
			}
		}
		
		PrintWriter log = new PrintWriter(Settings.PATH_OUTPUT + "_log");
		log.println("Logfile");
		log.println("~~~~~~~\n");
		
		long indexStartTime = System.currentTimeMillis();
		
		TripleSet ts = new TripleSet(Settings.PATH_TRAINING);
		PathSampler ps = new PathSampler(ts);		
		
		int pathCounter = 0;
		// int batchPathCounter = 0;
		int batchCounter = 0;


		int batchPreviouslyFoundRules = 0;
		int batchNewUsefulRules = 0;
		int batchRules = 0;

		boolean mineCyclicNotAcyclic = false;
		
		ArrayList<HashSet<Rule>> allUsefulRules = new ArrayList<HashSet<Rule>>();
		allUsefulRules.add(new HashSet<Rule>());
		
		
		int snapshotIndex = 0;
		
		int ruleSizeCyclic = 0;
		int ruleSizeAcyclic = 0;
		
		double lastCyclicCoverage = 0.0;
		double lastAcyclicCoverage = 0.0;
		
		long indexEndTime = System.currentTimeMillis();
		log.println("indexing dataset: " + Settings.PATH_TRAINING);
		log.println("time elapsed: " + (indexEndTime - indexStartTime) + "ms");
		log.println();
		log.println(IOHelper.getParams());
		log.flush();
		
		long startTime = System.currentTimeMillis();
		
		// Thread[] scorer = new Thread[Learn.WORKER_THREADS];
		
		Thread[] scorer = new Thread[Settings.WORKER_THREADS];
		for (int threadCounter = 0; threadCounter < Settings.WORKER_THREADS; threadCounter++) {
			System.out.println("* creating worker thread #" + threadCounter);
			scorer[threadCounter] = new Scorer(ts);
			scorer[threadCounter].start();
		}
		
		
		boolean done = false;
		while (true) {
			
			if (mineCyclicNotAcyclic && ruleSizeCyclic + 1 > Settings.MAX_LENGTH_CYCLIC) mineCyclicNotAcyclic = false;
			if (!mineCyclicNotAcyclic && ruleSizeAcyclic + 1 > Settings.MAX_LENGTH_ACYCLIC) mineCyclicNotAcyclic = true;
			
			int ruleSize = mineCyclicNotAcyclic ? ruleSizeCyclic : ruleSizeAcyclic;
			
			HashSet<Rule> usefulRules = allUsefulRules.get(ruleSize);
			long time = System.currentTimeMillis();
			int elapsedSeconds = (int)(time - startTime) / 1000;
			synchronized(Learn.class) {
				if (elapsedSeconds > Settings.SNAPSHOTS_AT[snapshotIndex] || done) {
					storeRules(Settings.PATH_OUTPUT, done ? 0 : Settings.SNAPSHOTS_AT[snapshotIndex], allUsefulRules, log, lastCyclicCoverage, lastAcyclicCoverage, elapsedSeconds);
					// writeLog(PATH_OUTPUT, SNAPSHOTS_AT[snapshotIndex], allUsefulRules);
					
					
					snapshotIndex++;
					if (!done) System.out.println("CREATED SNAPSHOT " + snapshotIndex + " after " + elapsedSeconds + " seconds");
					else System.out.println("CREATED FINAL SNAPSHOT 0 after " + elapsedSeconds + " seconds");
					if (snapshotIndex == Settings.SNAPSHOTS_AT.length || done) {
						log.flush();
						log.close();
						System.out.println("Bye, bye");
						Learn.finished = true;
						try {
							Thread.sleep(500);
						}
						catch (InterruptedException e) {
							e.printStackTrace();
						}
						System.exit(1);
					}
				}
			}
			long batchStartTime = System.currentTimeMillis();
			long currentTime = 0;
			

			
			while (true) {
			
				currentTime = System.currentTimeMillis();
				if (currentTime- batchStartTime > Settings.BATCH_TIME) break;
				pathCounter++;
				// takeTime();
				Path p = ps.samplePath(ruleSize + 2, mineCyclicNotAcyclic);
				if (p != null && p.isValid()) {
					
					// if (mineCyclicNotAcyclic && !p.isCyclic()) continue;
					// if (!mineCyclicNotAcyclic && p.isCyclic()) continue;
					
					Rule pr = new Rule(p);
					HashSet<Rule> rules = pr.getGeneralizations(mineCyclicNotAcyclic);					
				
					for (Rule r : rules) {
						
						if (r.isTrivial()) continue;
						// if (!mineCyclicNotAcyclic && r.isXYRule()) continue;
						
						batchRules++;
						//if (r.isXYRule()) batchCyclicRules++;
						//else batchAcyclicRules++;
						if (!usefulRules.contains(r)) {
							if (!addRuleToBeScored(r)) {
								try {
									// System.out.println(" ### sleep there are already too many rules to be scored ");
									Thread.sleep(25);
									// System.out.println("looper sleeps");
								}
								catch (InterruptedException e) {  e.printStackTrace(); }	
							}
							Rule scoredRule = getScoredRule();
							while (scoredRule != null) {
								if (scoredRule.getConfidence() >= Settings.THRESHOLD_CONFIDENCE &&  scoredRule.getCorrectlyPredicted() >= Settings.THRESHOLD_CORRECT_PREDICTIONS) {
									
									//if (r.isXYRule()) batchNewUsefulCyclicRules++;
									//else batchNewUsefulAcyclicRules++;
									if (usefulRules.add(scoredRule))  batchNewUsefulRules++;
								}
								scoredRule = getScoredRule();
							}
						}
						else batchPreviouslyFoundRules++;
					}
				}
			}
			synchronized(Learn.class) {
				batchCounter++;
				String type = mineCyclicNotAcyclic ? "CYCLIC" : "ACYCLIC";
				double currentCoverage = batchPreviouslyFoundRules / (double)(batchNewUsefulRules + batchPreviouslyFoundRules);
				System.out.println();
				System.out.println(">>> batch [" + type + " " + (ruleSize+1) +"] #" + batchCounter + " (sampled " + pathCounter + " pathes) *****");
				System.out.println(">>> fraction of previously seen rules within useful rules in this batch: " +  currentCoverage + "");
				System.out.println(">>> new rules above thresholds within this thread: " + batchNewUsefulRules);
				System.out.println(">>> rules found previously or a second time in this batch: " + batchPreviouslyFoundRules);
				System.out.println(">>> all rules created during within this batch: " + batchRules);
				System.out.println(">>> all rules stored so far: " + usefulRules.size());
				if (mineCyclicNotAcyclic) lastCyclicCoverage = currentCoverage;
				else lastAcyclicCoverage = currentCoverage;
				if (currentCoverage > Settings.SATURATION || batchNewUsefulRules == 0) {
					ruleSize++;
					if (mineCyclicNotAcyclic) ruleSizeCyclic = ruleSize;
					if (!mineCyclicNotAcyclic) ruleSizeAcyclic = ruleSize;
					System.out.println(">>> ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ ");
					System.out.println(">>> INCREASING RULE SIZE OF " + type + " RULE TO " + (ruleSize + 1));
					System.out.println(">>> ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
					log.println("increasing rule size of " + type + " rules to " + (ruleSize + 1) + " after " + ((startTime - System.currentTimeMillis()) / 1000) + "s");
					log.println();
					log.flush();
					allUsefulRules.add(new HashSet<Rule>());
				}
				
				batchNewUsefulRules = 0;
				batchRules = 0;
				batchPreviouslyFoundRules = 0;
	
				mineCyclicNotAcyclic = !mineCyclicNotAcyclic;
				if (mineCyclicNotAcyclic && ruleSizeCyclic + 1 > Settings.MAX_LENGTH_CYCLIC) {
					mineCyclicNotAcyclic = false;	
				}
				
				if (!mineCyclicNotAcyclic && ruleSizeAcyclic + 1 > Settings.MAX_LENGTH_ACYCLIC) {
					mineCyclicNotAcyclic = true;
				}
				
				if (ruleSizeCyclic + 1 > Settings.MAX_LENGTH_CYCLIC && ruleSizeAcyclic + 1 > Settings.MAX_LENGTH_ACYCLIC) {
					done = true;
				}
				Learn.rulesScored.clear();
				Learn.rulesToBeScored.clear();
			}
		}
		
		
		
	}
	
	
	public static synchronized Rule getNextRuleToBeScored()  {
		return rulesToBeScored.poll();
	}
	
	public static synchronized boolean addRuleToBeScored(Rule rule)  {
		if (rulesToBeScored.size() > 10000) {
			return false;
		}
		else {
			rulesToBeScored.add(rule);
			return true;
		}
	}
	
	
	public static synchronized boolean addScoredRule(Rule rule)  {
		rulesScored.add(rule);
		if (rulesScored.size() > 10000) {
			return false;
		}
		else {
			return true;
		}
	}
	
	public static synchronized Rule getScoredRule()  {
		return rulesScored.poll();
	}
	
	public static boolean finished() {
		return finished;
	}
	

	public static void showElapsedMoreThan(long duration, String message) {
		long now = System.currentTimeMillis();
		long elapsed = now - timeStamp;
		if (elapsed > duration) {
			System.err.println(message + " required " + elapsed + " millis!");
		}
		
	}

	public static void takeTime() {
		timeStamp = System.currentTimeMillis();
		
	}

	private static void storeRules(String path, int snapshotCounter, ArrayList<HashSet<Rule>> rules, PrintWriter log, double lastCC, double lastAC, int elapsedSeconds) {
		DecimalFormat df = new DecimalFormat("###.#");
		File ruleFile = new File(path + "-" + snapshotCounter);
		
		// very dirty
		int maxBodySize = 10;
		int[] acyclicCounter = new int[maxBodySize];
		int[] cyclicCounter = new int[maxBodySize];
		
		try {
			log.println();
			log.println("rule file: " + ruleFile.getPath());
			System.out.println(">>> Storing rules in file " + ruleFile.getPath());
			PrintWriter pw = new PrintWriter(ruleFile);
			for (HashSet<Rule> rrules : rules) {
				for (Rule r : rrules) {
					if (r.bodysize() < maxBodySize)
					if (r.isXYRule()) cyclicCounter[r.bodysize()-1]++;
					else acyclicCounter[r.bodysize()-1]++;
					pw.println(r);
				}
			}
			pw.flush();
			pw.close();
			log.print("cyclic: ");
			for (int i = 0; i< maxBodySize; i++) {
				if (cyclicCounter[i] == 0) break;
				log.print(cyclicCounter[i] + " | ");
			}
			log.println(df.format(lastCC * 100) + "%");
			log.print("acyclic: ");
			for (int i = 0; i< maxBodySize; i++) {
				if (acyclicCounter[i] == 0) break;
				log.print(acyclicCounter[i] + " | ");
			}
			log.println(df.format(lastAC * 100) + "%");
			log.println("time planned: " + snapshotCounter +  "s");
			log.println("time elapsed: " + elapsedSeconds + "s");
			log.println("");
			log.flush();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		
		
		
		
	}




}
